import os
from os.path import join, basename, dirname

def createuser(buildout_user='buildout'):
    "Creates a user account to run the buildout in"
    #keyname="buildout_dsa.%s"%(buildout_host)
    #if not os.path.exists(keyname):
    if True:
        sudo('test -d ~$(buildout_user) || adduser $(buildout_user)')
        sudo('test -d ~$(buildout_user)/.ssh || mkdir ~$(buildout_user)/.ssh;')
        sudo('(chmod 700 ~$(buildout_user)/.ssh; touch ~$(buildout_user)/.ssh/authorized_keys)')
        sudo('chmod 600 ~$(buildout_user)/.ssh/authorized_keys')
        #run("rm -f /tmp/buildout_dsa")
        #run("ssh-keygen -t dsa -N '' -f /tmp/buildout_dsa")
        #run('rm ~$(buildout_user)/.ssh/buildout_dsa.pub')
        #try:
        #    download('/tmp/buildout_dsa','buildout_dsa')
        #    download('/tmp/buildout_dsa.pub','buildout_dsa.pub')
        #except:
        #    pass
        sudo('cp ~$(buildout_user)/.ssh/authorized_keys ~$(buildout_user)/.ssh/authorized_keys.bak')
        sudo('cat /tmp/buildout_dsa.pub >> ~$(buildout_user)/.ssh/authorized_keys')
    set(fab_key_filename=keyname)


def predeploy(hostout):
    "install buildout and its dependencies"
    #run('export http_proxy=localhost:8123') # TODO get this from setting

    set(dist_dir = hostout.getDownloadCache(),
        unified='Plone-3.2.1r3-UnifiedInstaller',
        unified_url='http://launchpad.net/plone/3.2/3.2.1/+download/Plone-3.2.1r3-UnifiedInstaller.tgz',
        )

    sudo('mkdir -p $(dist_dir)/dist && sudo chown -R $(effectiveuser) $(dist_dir)')
    sudo('mkdir -p %(dc)s && sudo chown $(effectiveuser) %(dc)s'% dict(dc=hostout.getEggCache()))

    #Download the unified installer if we don't have it
    sudo('test -f $(buildout_dir)/bin/buildout || \
         test -f $(dist_dir)/$(unified).tgz || \
         ( cd /tmp && \
         wget  --continue $(unified_url) && \
         sudo mv /tmp/$(unified).tgz $(dist_dir)/$(unified).tgz && \
         sudo chown $(effectiveuser) $(dist_dir)/$(unified).tgz \
         ) \
         ')
    # untar and run unified installer
    sudo('test -f $(buildout_dir)/bin/buildout || \
          (cd /tmp && \
          tar -xvf $(dist_dir)/$(unified).tgz && \
          test -d /tmp/$(unified) && \
          cd /tmp/$(unified) && \
          sudo mkdir -p  $(install_dir) && \
          sudo ./install.sh --target=$(install_dir) --instance=$(instance) --user=$(effectiveuser) --nobuildout standalone && \
          sudo chown -R $(effectiveuser) $(install_dir)/$(instance))')

    for cmd in hostout.getPreCommands():
        sudo('sh -c "%s"'%cmd)


#TODO Need to hook into init.d
#http://www.webmeisterei.com/friessnegger/2008/06/03/control-production-buildouts-with-supervisor/
#cd /etc/init.d/
#ln -s INSTANCE_HOME/bin/supervisord project-supervisord
#ln -s INSTANCE_HOME/bin/supervisorctl project-supervisorctl
#update-rc.d project-supervisord defaults


def dodeploy(hostout):
    "deploy the package of changed cfg files"

    #need to send package. cycledown servers, install it, run buildout, cycle up servers

    for pkg in hostout.localEggs():
        tmp = join('/tmp', basename(pkg))
        tgt = join(hostout.getDownloadCache(), 'dist', basename(pkg))
        try:
            sudo('test -f %s'%tgt)
        except:
            put(pkg, tmp)
            sudo("mv %s %s"%(tmp,tgt))
            sudo('chown $(effectiveuser) %s' % tgt)

    package=hostout.getHostoutPackage()
    tmp = join('/tmp', basename(package))
    tgt = join(hostout.getDownloadCache(), basename(package))

    try:
        sudo('test -f %s'%tgt)
    except:
        put(package, tmp)
        sudo("mv %s %s"%(tmp,tgt))
        sudo('chown $(effectiveuser) %s' % tgt)

    set(
        #fab_key_filename="buildout_dsa",
        dist_dir=hostout.dist_dir,
        install_dir=hostout.remote_dir,
        hostout_file=hostout.getHostoutFile(),
    )

    #need a way to make sure ownership of files is ok
    sudo('tar --no-same-permissions --no-same-owner --overwrite --owner $(effectiveuser) -xvf %s --directory=$(install_dir)' % tgt)
#    if hostout.getParts():
#        parts = ' '.join(hostout.getParts())
 #       sudo('sudo -u $(effectiveuser) sh -c "cd $(install_dir) && bin/buildout -c $(hostout_file) install %s"' % parts)
  #  else:
    #Need to set home var for svn to work
    sudo('sudo -u $(effectiveuser) sh -c "export HOME=~$(effectiveuser) && cd $(install_dir) && bin/buildout -c $(hostout_file)"')

#    run('cd $(install_dir) && $(reload_cmd)')
#    sudo('chmod 600 .installed.cfg')
#    sudo('find $(install_dir)  -type d -name var -exec chown -R $(effectiveuser) \{\} \;')
#    sudo('find $(install_dir)  -type d -name LC_MESSAGES -exec chown -R $(effectiveuser) \{\} \;')
#    sudo('find $(install_dir)  -name runzope -exec chown $(effectiveuser) \{\} \;')


def deploy(hostout):
    ""
    if hostout.password:
        set(fab_password=hostout.password)
    if hostout.identityfile:
        set(fab_key_filename=hostout.identityfile)
    set(
        fab_user=hostout.user,
        fab_hosts=[hostout.host],
        effectiveuser=hostout.effective_user,
        buildout_dir=hostout.remote_dir,
        install_dir=os.path.split(hostout.remote_dir)[0],
        instance=os.path.split(hostout.remote_dir)[1],
        download_cache=hostout.getDownloadCache()
    )

    predeploy(hostout)
    dodeploy(hostout)
    postdeploy(hostout)

def postdeploy(hostout):
    for cmd in hostout.getPostCommands():
        sudo('sh -c "%s"'%cmd)





##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import logging, os, shutil, tempfile, urllib2, urlparse
import setuptools.archive_util
import datetime

import zc.buildout
import fabric
import tarfile
import ConfigParser
import sys
import md5
import os
from zc.buildout import buildout
from os.path import join, exists
from itertools import chain
import re
from zc.buildout.buildout import Buildout
from paramiko import DSSKey, PKey
from paramiko import SSHConfig

import time, random, md5
from collective.hostout import relpath
import pkg_resources


"""
1. ensure we are on trunk and up to date somehow.
1. Find any dependencies that need a new release and increment the version and create a distribution
1. create a hostout.cfg which is a repeatable buildout which is pinned for deployment by listing all
of all the eggs.
2. version this + all dev dependencies with a tag so can recover this version.
4. bundle the cfg up + eggs (maybe just dev eggs)
5. send to host
6. setup host it need be
7. overwrite with bundle and build
"""

def clean(lines):
    if lines is None:
        return []
    return [l.strip() for l in lines.split('\n') if l.strip() != '']

_isurl = re.compile('([a-zA-Z0-9+.-]+)://').match


def get_all_extends(cfgfile):
    if _isurl(cfgfile):
        return []

    config = ConfigParser.ConfigParser()
    config.read([cfgfile])
    files = [cfgfile]
    if not 'buildout' in config.sections():
        return files
    if not 'extends' in config.options('buildout'):
        return files
    extends = chain(*[el.split() for el in clean(config.get('buildout', 'extends'))])
    curdir = os.path.dirname(cfgfile)
    for extend in extends:
        if not _isurl(extend):
            extend = os.path.join(curdir, extend)
        files.extend(get_all_extends(extend))
    return files

class HostOut:
    def __init__(self, name, opt, packages):

        self.buildout_dir = packages.buildout_location
        self.dist_dir = packages.dist_dir
        self.packages = packages
        self.hostout_package = None

        self.name = name
        self.effective_user = opt['effective-user']
        self.remote_dir = opt['path']
        self.host = opt['host']
        self.user = opt['user']
        self.password = opt['password']
        self.identityfile = opt['identity-file']
        self.start_cmd = opt.get('post-commands')
        self.stop_cmd = opt.get('pre-commands')
        self.extra_config = opt['include']
        self.buildout_cfg = [p.strip() for p in opt['buildout'].split() if p.strip()]
        self.versions_part = opt.get('versions','versions')
        self.parts = [p.strip() for p in opt['parts'].split() if p.strip()]
        self.buildout_cache = opt.get('buildout-cache','')
        if not self.buildout_cache:
            install_base = os.path.dirname(self.getRemoteBuildoutPath())
            self.buildout_cache = os.path.join(install_base,'buildout-cache')

        from pkg_resources import resource_string, resource_filename
        fabfile = resource_filename(__name__, 'fabfile.py')

        self.fabfiles = [p.strip() for p in opt['fabfiles'].split() if p.strip()] + [fabfile]

        #self.packages = opt['packages']
        #dist_dir = os.path.abspath(os.path.join(self.buildout_location,self.dist_dir))
        #if not os.path.exists(dist_dir):
        #    os.makedirs(dist_dir)
        #self.tar = None

    def getHostoutFile(self):
        #make sure package has generated
        self.getHostoutPackage()
        return self.config_file[len(self.packages.buildout_location)+1:]

    def getPreCommands(self):
        return self._subRemote(clean(self.stop_cmd))

    def getPostCommands(self):
        return self._subRemote(clean(self.start_cmd))

    def getBuildoutDependencies(self):
        abs = lambda p: os.path.abspath(os.path.join(self.getLocalBuildoutPath(),p))
        return [abs(p) for p in clean(self.extra_config)]

    def getLocalBuildoutPath(self):
        return os.path.abspath(self.packages.buildout_location)

    def getRemoteBuildoutPath(self):
        return self.remote_dir

    def localEggs(self):
        self.getHostoutPackage() #ensure eggs are generated
        return self.packages.local_eggs

    def getParts(self):
        return self.parts

    def getDownloadCache(self):
        return "%s/%s" % (self.buildout_cache, 'downloads')
    def getEggCache(self):
        return "%s/%s" % (self.buildout_cache, 'eggs')

    def _subRemote(self, cmds):
        "replace abs localpaths to the buildout with absluote remate buildout paths"
        return [c.replace(self.getLocalBuildoutPath(), self.getRemoteBuildoutPath()) for c in cmds]

#    def getDeployTar(self):
#        return self.packages.getDeployTar()

    def getHostoutPackage(self):
        "determine all the buildout files that make up this configuration and package them"

        if self.hostout_package is not None:
            return self.hostout_package

        folder = self.dist_dir

        dist_dir = self.packages.dist_dir
        self.config_file = self.genhostout()
        config_file = os.path.abspath(os.path.join(self.packages.buildout_location,self.config_file))
        base = os.path.dirname(config_file)
        if not os.path.exists(config_file):
            raise "Invalid config file"

        files = get_all_extends(config_file)
        files += self.getBuildoutDependencies()

        self.packages.release_eggs()

        self.packages.writeVersions(config_file, self.versions_part)

        dist_dir = self.dist_dir
        self.releaseid = '%s_%s'%(time.time(),uuid())
        self.releaseid = _dir_hash(files)

        name = '%s/%s_%s.tgz'%(dist_dir,'deploy', self.releaseid)
        self.hostout_package = name
        if os.path.exists(name):
            return name
        else:
            self.tar = tarfile.open(name,"w:gz")

        for file in files:
            relative = file[len(self.buildout_dir)+1:] #TODO
            self.tar.add(file,arcname=relative)
        self.tar.close()
        return self.hostout_package


    def getDSAKey(self):
        keyfile = os.path.abspath(os.path.join(self.buildout_location,'hostout_dsa'))
        if not os.path.exists(keyfile):
            key = DSSKey.generate()
            key.write_private_key_file(keyfile)
        else:
            key = PKey.from_private_key_file(DSSKey, keyfile)
        return keyfile, str(key)

    def readsshconfig(self):
        config = os.path.expanduser('~/.ssh/config')
        if not os.path.exists(config):
            return
        f = open(config,'r')
        sshconfig = SSHConfig()
        sshconfig.parse(f)
        f.close()
        host = self.host
        try:
            host,port = host.split(':')
        except:
            port = None
        opt = sshconfig.lookup(host)

        if port is None:
            port = opt.get('port')

        host = opt.get('hostname', host)
        if port:
            host = "%s:%s" % (host,port)
        self.host=host
        if not self.identityfile:
            self.identityfile = opt.get('identityfile', None)
            if self.identityfile:
                self.identityfile = os.path.expanduser(self.identityfile).strip()
        if not self.user:
            self.user=opt.get('user','root')



    def runfabric(self, command):
        "return all commands if none found to run"

        cmds = {}
        cmd = None

        try:
            try:

                for fabfile in reversed(self.fabfiles):

                    fabric._load_default_settings()
                    fabric.load(fabfile, fail='warn')
                    cmds.update(fabric.COMMANDS)
                    cmd = fabric.COMMANDS.get(command, None)
                    if cmd is not None:
                        cmd(self)
                if cmd is None:
                    return cmds.keys()

            finally:
                fabric._disconnect()
            print("Done.")
        except SystemExit:
            # a number of internal functions might raise this one.
            raise
        except KeyboardInterrupt:
            print("Stopped.")
        #    except:
        #        sys.excepthook(*sys.exc_info())
        #        # we might leave stale threads if we don't explicitly exit()
        #        return False

#    def genhostout(self):
#        """ generate a new buildout file which pins versions and uses our deployment distributions"""
#

#        base = self.buildout_dir


#        files = [relpath(file, base) for file in self.buildout_cfg]
        #dist_dir = relpath(self.dist_dir, base)
        #versions = ""
#        hostout = HOSTOUT_TEMPLATE % dict(buildoutfile=' '.join(files),
                                          #eggdir=dist_dir,
 #                                         download_cache=self.getDownloadCache(),
 #                                         egg_cache=self.getEggCache(),
 #                                         )
 #       path = os.path.join(base,'%s.cfg'%self.name)
 #       hostoutf = open(path,'w')
 #       hostoutf.write(hostout)
 #       hostoutf.close()
 #       return path

    def genhostout(self):
        base = self.buildout_dir
        path = os.path.join(base,'%s.cfg'%self.name)
        config = ConfigParser.ConfigParser()
        config.read([path])
        if 'buildout' not in config.sections():
            config.add_section('buildout')
        files = [relpath(file, base) for file in self.buildout_cfg]

        config.set('buildout', 'extends', ' '.join(files))
        config.set('buildout', 'develop', '')
        config.set('buildout', 'eggs-directory', self.getEggCache())
        config.set('buildout', 'download-cache', self.getDownloadCache())
        config.set('buildout', 'newest', 'true')
        if self.getParts():
            config.set('buildout', 'parts', ' '.join(self.getParts()))

        fp = open(path,'w')
        config.write(fp)
        fp.close()
        return path



HOSTOUT_TEMPLATE = """
[buildout]
extends = %(buildoutfile)s

#prevent us looking for them as developer eggs
develop=

#install-from-cache = true

#Match to unifiedinstaller
eggs-directory = %(egg_cache)s
download-cache = %(download_cache)s

#non-newest set because we know exact versions we want
newest=true
"""



import zc.buildout.easy_install
from zc.buildout.buildout import pkg_resources_loc


class Packages:
    """ responsible for packaging the development eggs ready to be released to each host"""

    def __init__(self, config):
        self.packages = packages = [p for p in config.get('buildout','packages').split()]

        self.buildout_location = config.get('buildout', 'location')
        self.dist_dir = config.get('buildout','dist_dir')
        self.versions = dict(config.items('versions'))
        self.tar = None
        dist_dir = os.path.abspath(os.path.join(self.buildout_location,self.dist_dir))
        if not os.path.exists(dist_dir):
            os.makedirs(dist_dir)
        self.dist_dir = dist_dir
        self.local_eggs = []

    def getDistEggs(self):

        res = {}

        localdist_dir = self.dist_dir
        eggs = pkg_resources.find_distributions(localdist_dir)
        return dict([(( egg.project_name,egg.version),egg) for egg in eggs])


    def release_eggs(self):
        "developer eggs->if changed, increment versions, build and get ready to upload"
        # first get list of deveelop packages we got from recipe
        # for each package
        #
        if self.local_eggs:
            return self.local_eggs

        #python setup.py sdist bdist_egg
 #       tmpdir = tempfile.mkdtemp()
        localdist_dir = self.dist_dir
        eggs = self.getDistEggs()

        donepackages = []
        ids = {}
        self.develop_versions = {}
        released = {}
        if self.packages:
            print "Hostout: Preparing eggs for transport"
        for path in self.packages:

            # use buildout to run setup for us
            hash = _dir_hash([path])
            ids[hash]=path
            dist = [d for d in pkg_resources.find_distributions(path, only=True)]
            if dist:
                dist = dist[0]
                egg = eggs.get( (dist.project_name, dist.version) )
            else:
                egg = None
            if egg is not None and hash in dist.version:
                self.develop_versions[dist.project_name] = dist.version
                self.local_eggs.append(egg.location)
            elif os.path.isdir(path):
                print "Hostout: Develop egg %s changed. Releasing with hash %s" % (path,hash)
                res = self.setup(args=[path,
                                     'clean',
                                     'egg_info',
                                     '--tag-build','dev_'+hash,
                                     #'sdist',
                                     #'--formats=zip', #fix bizzare gztar truncation on windows
                                      'bdist_egg',
                                     '--dist-dir',
                                     '%s'%localdist_dir,
                                      ])
                dist = [d for d in pkg_resources.find_distributions(path, only=True)]
                dist = dist[0]
                self.develop_versions[dist.project_name] = dist.version
                released[dist.project_name] = dist.version
            else:
#                shutil.copy(path,localdist_dir)
                self.local_eggs.append(path)
        if released:
            eggs = self.getDistEggs()
            for (name,version) in released.items():
                egg = eggs.get( (name, version) )
                if egg is not None:
                    self.local_eggs.append(egg.location)
                else:
                    raise "Egg wasn't generated. See errors above"


        if self.local_eggs:
            specs = ["\t%s = %s"%p for p in self.develop_versions.items()]
            print "Hostout: Eggs to transport:\n%s" % '\n'.join(specs)
        return self.local_eggs

    def getVersion(self, path):
        "Test to see if we already have a release of this developer egg"
        dist = [d for d in pkg_resources.find_distributions(path, only=True)]
        dist = dist[0]

        return dist.version

    def writeVersions(self, versions_file, part):

#        assert len(specs) == len(self.packages)
        config = ConfigParser.ConfigParser()
        config.read([versions_file])
        specs = {}
        specs.update(self.versions)
        specs.update(self.develop_versions)
        config.set('buildout', 'versions', part)
        if part not in config.sections():
            config.add_section(part)
        for name, version in specs.items():
            config.set(part,name,version)
        fp = open(versions_file,'w')
        config.write(fp)
        fp.close()
        print "Hostout: Wrote versions to %s"%versions_file


    def setup(self, args):
        setup = args.pop(0)
        if os.path.isdir(setup):
            setup = os.path.join(setup, 'setup.py')

        #self._logger.info("Running setup script %r.", setup)
        setup = os.path.abspath(setup)

        fd, tsetup = tempfile.mkstemp()
        try:
            os.write(fd, zc.buildout.easy_install.runsetup_template % dict(
                setuptools=pkg_resources_loc,
                setupdir=os.path.dirname(setup),
                setup=setup,
                __file__ = setup,
                ))
            os.spawnl(os.P_WAIT, sys.executable, zc.buildout.easy_install._safe_arg (sys.executable), tsetup,
                      *[zc.buildout.easy_install._safe_arg(a)
                        for a in args])
        finally:
            os.close(fd)
            os.remove(tsetup)


def main(cfgfile, args):
    "execute the fabfile we generated"

    config = ConfigParser.ConfigParser()
    config.read([cfgfile])
    files = [cfgfile]
    allhosts = {}
#    buildout = Buildout(config.get('buildout','buildout'),[])
    packages = Packages(config)
    #eggs = packages.release_eggs()
    for section in [s for s in config.sections() if s not in ['buildout', 'versions']]:
        options = dict(config.items(section))

        hostout = HostOut(section, options, packages)
        allhosts[section] = hostout
    if args:
        cmd, hosts = args[0],args[1:]
        if 'all' in hosts:
            torun = allhosts.values()
        else:
            torun = [allhosts[host] for host in hosts if host in allhosts]
        if not torun:
            print >> sys.stderr, "Invalid hostout hostouts are: %s"% ' '.join(allhosts.keys())
        elif cmd:
            for hostout in torun:
                hostout.readsshconfig()
                res = hostout.runfabric(cmd)
                if res:
                    print >> sys.stderr, "Invalid command. Valid commands are - %s"%res

    else:
        print >> sys.stderr, "Please specify a command"



def uuid( *args ):
  """
    Generates a universally unique ID.
    Any arguments only create more randomness.
  """
  t = long( time.time() * 1000 )
  r = long( random.random()*100000000000000000L )
  try:
    a = socket.gethostbyname( socket.gethostname() )
  except:
    # if we can't get a network address, just imagine one
    a = random.random()*100000000000000000L
  data = str(t)+' '+str(r)+' '+str(a)+' '+str(args)
  data = md5.md5(data).hexdigest()
  return data

ignore_directories = '.svn', 'CVS', 'build', '.git'
ignore_files = ['PKG-INFO']
def _dir_hash(paths):
    hash = md5.new()
    for path in paths:
        if os.path.isdir(path):
            walked = os.walk(path)
        else:
            walked = [(os.path.dirname(path), [], [os.path.basename(path)])]
        for (dirpath, dirnames, filenames) in walked:
            dirnames[:] = [n for n in dirnames if not (n in ignore_directories or n.endswith('.egg-info'))]
            filenames[:] = [f for f in filenames
                        if not (f in ignore_files or f.endswith('pyc') or f.endswith('pyo'))]
            hash.update(' '.join(dirnames))
            hash.update(' '.join(filenames))
            for name in filenames:
                hash.update(open(os.path.join(dirpath, name)).read())
    import base64
    hash = base64.urlsafe_b64encode(hash.digest()).strip()
    hash = hash.replace('_','-').replace('=','')
    return hash





##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os, re, StringIO, sys, tarfile
import zc.buildout.testing

import unittest
import doctest
import zope.testing
from zope.testing import renormalizing
from zc.buildout.tests import easy_install_SetUp
from zc.buildout.tests import normalize_bang
import os
from socket import socket
from paramiko import Transport, ServerInterface, AUTH_SUCCESSFUL, OPEN_SUCCEEDED
import paramiko as ssh
from threading import Event, Thread


current_dir = os.path.abspath(os.path.dirname(__file__))
recipe_location = current_dir

for i in range(2):
    recipe_location = os.path.split(recipe_location)[0]


def setUp(test):
    #zc.buildout.tests.easy_install_SetUp(test)
    zc.buildout.testing.buildoutSetUp(test)
    zc.buildout.testing.install_develop('collective.hostout', test)
    zc.buildout.testing.install('functools', test)
    zc.buildout.testing.install('Fabric<0.1.0', test)
    zc.buildout.testing.install('paramiko', test)
    zc.buildout.testing.install('pycrypto', test)
    zc.buildout.testing.install('zc.recipe.egg', test)


class LocalSSH(ServerInterface, Thread):
    def __init__(self):
        Thread.__init__(self)
        self.socket = socket()
        self.socket.bind(('127.0.0.1', 10022))
        self.socket.listen(4)
        self.key = ssh.RSAKey.generate(160)

    def run(self):
        s,addr = self.socket.accept()
        t = Transport(s)
        t.add_server_key(self.key)
        e = Event()
        t.start_server(e, server=self)

    def get_allowed_auths(self, username):
        return "none,password,publickey"
    def check_auth_none(self, username):
        return AUTH_SUCCESSFUL
    def check_auth_password(self, username, password):
        return AUTH_SUCCESSFUL
    def check_auth_publickey(self, username, key):
        return AUTH_SUCCESSFUL
    def check_channel_request(self, kind, chanid):
        return OPEN_SUCCEEDED
    def check_channel_exec_request(self, channel, command):
        return True
    def check_channel_shell_request(self, channel):
            self.channel = channel
            self.fs = channel.makefile()
            return True
    def read(self):
        return self.fs.read()


localssh = LocalSSH()
localssh.start()
client  = ssh.SSHClient()
client.connect('localhost', 10022, 'root')


def add(tar, name, src, mode=None):
    info.size = len(src)
    if mode is not None:
        info.mode = mode
    tar.addfile(info, StringIO.StringIO(src))



def test_suite():

    globs = globals()


    return unittest.TestSuite((
        #doctest.DocTestSuite(),
        doctest.DocFileSuite(
            'README.txt',
             package='collective.hostout',
            setUp=setUp, tearDown=zc.buildout.testing.buildoutTearDown,
            optionflags = doctest.ELLIPSIS | doctest.REPORT_ONLY_FIRST_FAILURE |
                        doctest.NORMALIZE_WHITESPACE,
                        globs=globs,
            checker=renormalizing.RENormalizing([
               zc.buildout.testing.normalize_path,
               #zc.buildout.testing.normalize_script,
               #zc.buildout.testing.normalize_egg_py,
               #zc.buildout.tests.normalize_bang,
               ]),
            ),


        ))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import logging, os, shutil, tempfile, urllib2, urlparse
import setuptools.archive_util
import datetime
import sha
import shutil
import zc.buildout
import zc.recipe.egg
from os.path import join
import os
from os.path import dirname, abspath
import ConfigParser


def system(c):
    if os.system(c):
        raise SystemError("Failed", c)



class Recipe:

    def __init__(self, buildout, name, options):
        self.egg = zc.recipe.egg.Egg(buildout, options['recipe'], options)
        self.name, self.options, self.buildout = name, options, buildout

        #get all recipes here to make sure we're the last called
        main = None
        for part, recipe, opt in self.getAllRecipes():
            if recipe == self.options['recipe']:
                main = opt
                if opt.get('mainhostout') == self.name:
                    break
        if main:
            main['mainhostout'] = self.name


        self.buildout_dir = self.buildout.get('buildout').get('directory')
        self.download_cache = self.buildout['buildout'].get('download-cache')
        self.install_from_cache = self.buildout['buildout'].get('install-from-cache')
        self.options['versions'] = self.buildout['buildout'].get('versions','versions')
        self.options['location'] = os.path.join(
            self.buildout['buildout']['parts-directory'],
            'hostout',
            )
        self.optionsfile = join(self.options['location'],'hostout.cfg')
        self.fabfiles = []

        extends = [e.strip() for e in self.options.get('extends','').split() if e.strip()]

        for extension in reversed(extends):
            extension = buildout.get(extension)
            if extension is None:
                continue
            for key in extension:
                if key == 'fabfile':
                    self.fabfiles.append(extension[key])
                if key not in self.options:
                    self.options[key] = extension[key]


        self.options.setdefault('dist_dir','dist')
        self.options.setdefault('buildout','buildout.cfg')
        self.options.setdefault('user','') #get default from .ssh/config
        self.options.setdefault('identity-file',self.options.get('identity_file',''))
        self.options.setdefault('effective-user','plone')
        self.options.setdefault('host', 'localhost')
        self.options.setdefault('password','')
        self.options.setdefault('post-commands', self.options.get('start_cmd',''))
        self.options.setdefault('pre-commands', self.options.get('stop_cmd', ''))
        self.options.setdefault('include', self.options.get('extra_config',''))
        self.options.setdefault('parts','')
        default_path = '~%s/buildout'%self.options['user']
        self.options.setdefault('path', self.options.get('remote_path',''))
#        self.extra_config = [s.strip() for s in self.options.get('extra_config','').split('\n') if s.strip()]
        self.options.setdefault('buildout_location',self.buildout_dir)
        self.options['fabfiles'] = '\n\t'.join(self.fabfiles)



    def install(self):
        logger = logging.getLogger(self.name)

        location = self.options['location']

        if not os.path.exists(location):
            os.mkdir(location)


        config = ConfigParser.ConfigParser()
        config.read(self.optionsfile)

        fp = open(self.optionsfile, 'w+')
        config.write(fp)
        fp.close()
        self.update()

        return self.optionsfile

    def update(self):

        if self.options.get('mainhostout') is not None:
            requirements, ws = self.egg.working_set()
            bin = self.buildout['buildout']['bin-directory']
            extra_paths=[]
            zc.buildout.easy_install.scripts(
                [('hostout', 'collective.hostout.hostout', 'main')],
                ws, self.options['executable'], bin,
                arguments="'%s',sys.argv[1:]"%self.optionsfile,
                extra_paths=extra_paths
#                initialization=address_info,
                )

        config = ConfigParser.ConfigParser()
        config.read(self.optionsfile)

        if not config.has_section(self.name):
            config.add_section(self.name)
        if not config.has_section('buildout'):
            config.add_section('buildout')
        for name,value in self.options.items():
            config.set(self.name, name, value)
        config.set('buildout','location',self.buildout_dir)

        if not config.has_section('versions'):
            config.add_section('versions')
        if self.options.get('mainhostout') is not None:
            for pkg,info in self.getVersions().items():
                    version,deps = info
                    config.set('versions',pkg,version)
            config.set('buildout', 'bin-directory', self.buildout.get('buildout').get('directory'))
            if self.options['dist_dir']:
                config.set('buildout','dist_dir', self.options['dist_dir'])

            packages = [p.strip() for p in self.buildout.get('buildout').get('develop','').split()]
            packages += [p.strip() for p in self.options.get('packages','').split()]
            if not config.has_section('buildout'):
                config.add_section('buildout')
            config.set('buildout', 'packages', '\n   '.join(packages))
        #self.options.setdefault('develop','')

        fp = open(self.optionsfile, 'w+')
        config.write(fp)
        fp.close()
        return self.optionsfile



    def getAllRecipes(self):
        recipes = []

        for part in [p.strip() for p in self.buildout['buildout']['parts'].split()]:
            options = self.buildout.get(part) #HACK
            if options is None or not options.get('recipe'):
                continue
            try:
                recipe,subrecipe = options['recipe'].split(':')
            except:
                recipe=options['recipe']
            recipes.append((part,recipe,options))
        return recipes

    def getVersions(self):
        versions = {}
        for part, recipe, options in self.getAllRecipes():
            egg = zc.recipe.egg.Egg(self.buildout, recipe, options)
            #TODO: need to put in recipe versions too
            requirements, ws = egg.working_set()
            for dist in ws.by_key.values():
                project_name =  dist.project_name
                version = dist.version
                old_version,dep = versions.get(project_name,('',[]))
                if recipe not in dep:
                    dep.append(recipe)
                if version != '0.0':
                    versions[project_name] = (version,dep)
        spec = ""
        return versions
        for project_name,info in versions.items():
            version,deps = info
            spec+='\n'
#            for dep in deps:
#                spec+='# Required by %s==%s\n' % (dep, 'Not Implemented') #versions[dep][0])
            if version != '0.0':
                spec+='%s = %s' % (project_name,version)+'\n'
#            else:
#                spec+='#%s = %s' % (project_name,version)+'\n'
        return spec


# relpath.py
# R.Barran 30/08/2004

import os

def relpath(target, base=os.curdir):
    """
    Return a relative path to the target from either the current dir or an optional base dir.
    Base can be a directory specified either as absolute or relative to current dir.
    """

    if not os.path.exists(target):
        raise OSError, 'Target does not exist: '+target

    if not os.path.isdir(base):
        raise OSError, 'Base is not a directory or does not exist: '+base

    base_list = (os.path.abspath(base)).split(os.sep)
    target_list = (os.path.abspath(target)).split(os.sep)

    # On the windows platform the target may be on a completely different drive from the base.
    if os.name in ['nt','dos','os2'] and base_list[0] <> target_list[0]:
        raise OSError, 'Target is on a different drive to base. Target: '+target_list[0].upper()+', base: '+base_list[0].upper()

    # Starting from the filepath root, work out how much of the filepath is
    # shared by base and target.
    for i in range(min(len(base_list), len(target_list))):
        if base_list[i] <> target_list[i]: break
    else:
        # If we broke out of the loop, i is pointing to the first differing path elements.
        # If we didn't break out of the loop, i is pointing to identical path elements.
        # Increment i so that in all cases it points to the first differing path elements.
        i+=1

    rel_list = [os.pardir] * (len(base_list)-i) + target_list[i:]
    return os.path.join(*rel_list)




##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

name = "collective.hostout"
setup(
    name = name,
    version = "0.9",
    author = "Dylan Jay",
    author_email = "software@pretaweb.com",
    description = "one click deployment for buildout based applications",
    license = "GPL",
    keywords = "buildout, deploy, deployment, server, plone, django, host, hosting",
    url='http://www.python.org/pypi/'+name,
    long_description=(
        read('README.txt')
        + '\n' +
        read('CHANGES.txt')
        + '\n' +
        'Detailed Documentation\n'
        '**********************\n'
        + '\n' +
        read('collective', 'hostout', 'README.txt')
        + '\n'
        ),

    packages = find_packages(),
    include_package_data = True,
#    data_files = [('.', ['*.txt'])],
#    package_data = {'':('*.txt')},
    namespace_packages = ['collective'],
    install_requires = ['zc.buildout',
                        'zc.recipe.egg',
                        'setuptools',
                        'Fabric<0.1.0', #in order to make it 2.4 compatible
                        'functools' ,#needed for fabric to make it 2.4 compatible
                        ],
    entry_points = {'zc.buildout':
                    ['default = collective.hostout:Recipe']},
    zip_safe = False,
    )
